#\documentclass{beamer} % 2>/dev/null || <<'% build'
#
#%_____________________________________________________
#%
#%   The following command builds the whole project:
#%
#%   bash Binom_Bijection.tex
#%_____________________________________________________
#%
#
#\usepackage[english]{babel}
#\usepackage[utf8]{inputenc}
#\usepackage{hyperref}
#\usepackage{amsmath}
#\usepackage{amssymb}
#\usepackage{xcolor}
#\usepackage{listings}
#    \lstset{basicstyle=\scriptsize}
#    \lstset{showstringspaces=false}
#    \lstset{frame=shadowbox}
#    \lstset{rulesepcolor=\color{gray}}
#
#\newcommand{\ignore}[1]{}
#
#\mode<presentation>{\usetheme{Singapore}}
#\setbeamercovered{dynamic}
#\AtBeginSection[]{
#    \begin{frame}
#        \frametitle{Overview}
#        \tableofcontents[currentsection]
#    \end{frame}
#}
#
#\begin{document}
#
#\title{Binom Bijection \\
#       Storing k-Subsets Efficiently}
#\author{Volker Grabsch}
#\date{September 7, 2009}
#\begin{frame}
#    \titlepage
#\end{frame}
#
#\begin{frame}
#    \frametitle{Draft}
#    \begin{center}
#    This presentation is still a draft!
#    \end{center}
#\end{frame}
#
#\begin{frame}
#    \frametitle{Overview}
#    \tableofcontents
#\end{frame}
#
#\section{Motivation}
#
#\begin{frame}[<+->]
#\frametitle{Storing Numbers}
#Store 100 numbers $(a_0,\ldots,a_{99})$ of range $\{0,1,2\}$ efficiently!
#
#\bigskip
#
#\begin{itemize}
#\item Store each number in 2 bits \\
#      $\Rightarrow 2 \cdot 100 =$ \textbf{200 bits}
#\item Store one big number (100 digits, base 3)
#\begin{align*}
#f(a_0,\ldots,a_{99}) &= \sum_{i=0}^{99} 3^i a_i
#                     &\in \left\{ 0,\ldots,3^{100}-1 \right\}
#\end{align*}
#$\Rightarrow \left\lceil \log_2\left(3^{100}\right) \right\rceil =$ \textbf{159 bits}
#
#\bigskip
#
#\item $3^{100}$ states to encode $\Rightarrow$ impossible to do better
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->]
#\frametitle{Storing Permutations}
#Store a permutation of 100 numbers $(0,\ldots,99)$ efficiently!
#
#\bigskip
#
#\begin{itemize}
#\item Store each number in 7 bits \\
#      $\Rightarrow 7 \cdot 100 =$ \textbf{700 bits}
#\item Store one big number (100 digits, base 100) \\
#      $\Rightarrow \left\lceil \log_2\left(100^{100}\right) \right\rceil =$ \textbf{665 bits}
#\item Store one big multibase number \\
#      (99 digits, bases 100,99,\ldots,2) \\
#      $\Rightarrow \left\lceil \log_2\left(100!\right) \right\rceil =$ \textbf{525 bits}
#
#\bigskip
#
#\item $100!$ states to encode $\Rightarrow$ impossible to do better
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->]
#\frametitle{Storing $k$-Subsets}
#Store a choice of 30 out of 100 numbers $\{0,\ldots,99\}$ efficiently!
#
#\bigskip
#
#\begin{itemize}
#\item Store each number in 7 bits \\
#      $\Rightarrow 7 \cdot 30 =$ \textbf{210 bits}
#\item Store one big number (30 digits, base 100) \\
#      $\Rightarrow \left\lceil \log_2\left(100^{30}\right) \right\rceil =$ \textbf{200 bits}
#\item Store one big number (30 digits, base 71) \\
#      $\Rightarrow \left\lceil \log_2\left(71^{30}\right) \right\rceil =$ \textbf{185 bits}
#\item Store in classic set representation \\
#      $\Rightarrow$ \textbf{100 bits} (of which 30 are set to 1)
#
#\bigskip
#
#\item Theoretical minimum: \\
#      $\left\lceil \log_2\binom{100}{30} \right\rceil =$ \textbf{85 bits}
#\item Need a bijection
#\begin{equation*}
#\varphi : \binom{\{0,\ldots,99\}}{30}
#          \stackrel{\sim}{\rightarrow}
#          \left\{0,\ldots,\binom{100}{30}-1\right\}
#\end{equation*}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->]
#\frametitle{Binom Bijection}
#\begin{itemize}
#\item Question:
#\begin{align}
#\varphi : \binom{\{0,\ldots,n-1\}}{k}
#          &\stackrel{\sim}{\rightarrow}
#           \left\{0,\ldots,\binom{n}{k}-1\right\}
#\end{align}
#\item Solution:
#\begin{align}
#\varphi\left(\left\{s_0 < \ldots < s_{k-1}\right\}\right)
#    &= \sum_{i=0}^{k-1} \binom{s_i}{i+1}
#\end{align}
#\item Need
#    \begin{itemize}
#    \item constructive proof of bijectivity
#    \item implementation of $\varphi$
#    \item implementation of $\varphi^{-1}$
#    \end{itemize}
#\end{itemize}
#\end{frame}
#
#\section{Constructive Proof}
#
#\begin{frame}[<+->]
#\frametitle{Proof}
#\begin{itemize}
#\item Note: $\varphi$ is independent of $n$
#\item ...
#\begin{align*}
#\binom{n}{k} &= \binom{k-1}{k-1} + \cdots + \binom{n-1}{k-1} \\
#             &= \sum_{i=k-1}^{n-1} \binom{i}{k-1}
#\end{align*}
#\end{itemize}
#\end{frame}
#
#\section{Implementation}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Implementation of $\varphi$}
#\begin{itemize}
#\item Direct implementation of $\varphi$
#\begin{align*}
#\varphi\left(\left\{s_0 < \ldots < s_{k-1}\right\}\right)
#    &= \sum_{i=0}^{k-1} \binom{s_i}{i+1}
#\end{align*}
#\begin{lstlisting}[language=sh] % Binom_Bijection.py

def number_of_set(s):
    return sum(binom(si, i+1)
               for i, si in enumerate(sorted(s)))

#\end{lstlisting}
#\item Calculate binomial coefficients $\binom{n}{k}$ in Python
#\begin{lstlisting}[language=sh] % Binom_Bijection.py

import scipy

def binom(n, k):
    return scipy.comb(n, k, exact=1)

#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Naive implementation of $\varphi^{-1}$}
#\begin{itemize}
#\item Naive and slow implementation of $\varphi^{-1}$
#\begin{lstlisting}[language=sh] % Binom_Bijection.py

def set_of_number_SLOW(k, num):
    s = set()
    for i in xrange(k-1, -1, -1):
        si = i
        while num >= binom(si + 1, si - i):
            si += 1
        num -= binom(si, si - i - 1)
        s.add(si)
    return s

#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Implementation of $\varphi^{-1}$}
#\begin{itemize}
#\item Use previous binomial coefficient to calculate next one
#\begin{align*}
#\binom{n}{k} &= \frac{n}{k}\cdot\binom{n-1}{k-1}
#\end{align*}
#\item Fast implementation of $\varphi^{-1}$
#\begin{lstlisting}[language=sh] % Binom_Bijection.py

def set_of_number(k, num):
    s = set()
    for i in xrange(k-1, -1, -1):
        si = i
        b = 1
        while num >= b:
            si += 1
            b = (b * (si+1)) / (si-i)
        num -= (b * (si-i)) / (si+1)
        s.add(si)
    return s

#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Complexity}
#\begin{itemize}
#\item \verb|number_of_set|
#    \begin{itemize}
#    \item ...
#    \end{itemize}
#\item \verb|set_of_number|
#    \begin{itemize}
#    \item ...
#    \end{itemize}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Tests}
#\begin{itemize}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#from Binom_Bijection import number_of_set, set_of_number
#
#\end{lstlisting}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#assert number_of_set([]) == 0
#assert set_of_number(0, 0) == set([])
#
#\end{lstlisting}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#for s0 in xrange(0, 400):
#    assert number_of_set([s0]) == s0
#    assert set_of_number(1, s0) == set([s0])
#
#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Tests}
#\begin{itemize}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#counter = 0
#for s0 in xrange(1, 50):
#    for s1 in xrange(0, s0):
#        s = set([s1, s0])
#        assert number_of_set(s) == counter
#        assert set_of_number(2, counter) == s
#        counter += 1
#
#\end{lstlisting}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#counter = 0
#for s0 in xrange(2, 25):
#    for s1 in xrange(1, s0):
#        for s2 in xrange(0, s1):
#            s = set([s2, s1, s0])
#            assert number_of_set(s) == counter
#            assert set_of_number(3, counter) == s
#            counter += 1
#
#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Tests}
#\begin{itemize}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#def test_k(k, max):
#    def test_k_rec(s, i, max, counter):
#        if i == 0:
#            assert number_of_set(s) == counter[0]
#            assert set_of_number(k, counter[0]) == s
#            counter[0] += 1
#        else:
#            for si in xrange(i-1, max):
#                test_k_rec(s.union([si]), i-1, si,
#                           counter)
#    test_k_rec(set([]), k, max, [0])
#
#\end{lstlisting}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#test_k(4, 18)
#test_k(5, 14)
#test_k(6, 13)
#test_k(7, 13)
#test_k(14, 17)
#
#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Tests}
#\begin{itemize}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#def test_consistency(s):
#    assert set_of_number(len(s), number_of_set(s)) == s
#
#\end{lstlisting}
#\item ...
#\begin{lstlisting}[language=sh] % Tests.py
#
#test_consistency(set(xrange(0, 400)))
#test_consistency(set(xrange(0, 400, 2)))
#test_consistency(set(xrange(0, 400, 3)))
#test_consistency(set(xrange(0, 400, 10)))
#test_consistency(set(xrange(40, 400)))
#test_consistency(set(xrange(150, 400)))
#test_consistency(set(xrange(150, 400, 2)))
#test_consistency(set(xrange(150, 400, 3)))
#test_consistency(set(xrange(150, 400, 10)))
#test_consistency(set(xrange(200, 400)))
#test_consistency(set(xrange(360, 400)))
#\end{lstlisting}
#\end{itemize}
#\end{frame}
#
#\section{About this document}
#
#\begin{frame}[<+->][fragile]
#\frametitle{About this document}
#\begin{itemize}
#\item Website
#\begin{center}
#\url{http://www.profv.de/Binom_Bijection/}
#\end{center}
#
#\bigskip
#
#\item LaTeX source is also shell script
#\begin{lstlisting}[language=sh]%
#
#sh Binom_Bijection.tex
#
#\end{lstlisting}
#(\href{http://www.profv.de/literate-programming/}{Self-contained literate programming})
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Build instructions (I)}
#\begin{itemize}
#\item Shell options to handle errors correctly
#\begin{lstlisting}[language=sh]%
#% build
#
#set -euvx
#
#\end{lstlisting} % 2>/dev/null || <<'% build'
#\item Extract Python library
#\begin{lstlisting}[language=sh]%
#% build
#
#sed '/^[\%]/,/% Binom_Bijection.py$/ s,^,#,' \
#< Binom_Bijection.tex > Binom_Bijection.py
#
#\end{lstlisting} % 2>/dev/null || <<'% build'
#\item Extract and run tests
#\begin{lstlisting}[language=sh]%
#% build
#
#sed '/^[\%]/,/% Tests.py$/ s,^,#,' \
#< Binom_Bijection.tex > Tests.py
#
#python Tests.py
#
#\end{lstlisting} % 2>/dev/null || <<'% build'
#\end{itemize}
#\end{frame}
#
#\begin{frame}[<+->][fragile]
#\frametitle{Build instructions (II)}
#\begin{itemize}
#\item Compile LaTeX source to PDF
#\begin{lstlisting}[language=sh]%
#% build
#
#rm -f    Binom_Bijection.{aux,log,nav,out,snm,toc,vrb}
#pdflatex Binom_Bijection.tex
#pdflatex Binom_Bijection.tex
#rm -f    Binom_Bijection.{aux,log,nav,out,snm,toc,vrb}
#
#\end{lstlisting} % 2>/dev/null || <<'% build'
#\item Create ZIP archive
#\begin{lstlisting}[language=sh]%
#% build
#
#rm -rf  Binom_Bijection
#mkdir   Binom_Bijection
#cp      Binom_Bijection.{tex,pdf,py} Binom_Bijection/
#rm -f   Binom_Bijection.zip
#zip -r9 Binom_Bijection.zip Binom_Bijection/
#rm -rf  Binom_Bijection
#
#\end{lstlisting} % 2>/dev/null || <<'% build'
#\item Remove temporary files
#\begin{lstlisting}[language=sh]%
#% build
#
#rm -f Binom_Bijection.{py,pyc} Tests.py
#
#\end{lstlisting} % 2>/dev/null || <<'% build'
#\end{itemize}
#\end{frame}
#
#\end{document}
